package org.signwriting;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.MissingResourceException;

/**
 * reads the symbol file
 */
public class Dictionary {

    public static final int SWDICT_VERSION = 2;
    public static final int NUM_VOLUMES = 2;
    public static final int MAX_DE = 5399;
    public static final int MAX_DK = 9;

    private static final int MAX_PAIR = 4;

    String filename;
    //InitDict
    short dheader[] = new short[8];
    
    private Resources resources;

    int numEntries;
    DictionaryEntry[] entries;
    int dictFree;
    int dictRecs;
    DictionaryFile dictFile;

    private static Dictionary dictionary;
    private boolean failed = false;

    /**
     * constructor, takes a file name, opens it and reads it
     */
    private Dictionary() {
        resources = Resources.getResources();
        filename = resources.getString("Dictionary.prefix");
        try {
            read();
        } catch (IOException ioe) {
            com.dtai.util.ShowUser.error(ioe,"Error reading dictionary "+filename);
            failed = true;
        }
    }
    
    public static Dictionary getDictionary() {
        if (dictionary == null) {
            dictionary = new Dictionary();
        }
        return dictionary;
    }
    
    public String getSignLabel(String resourceKey)
        throws MissingResourceException {
        if (failed) {
            throw new MissingResourceException("Dictionary Failure",
                                        "ResourceBundle", resourceKey);
        }
        getDictionary();
        resourceKey = resourceKey+".sign";
        String dictionaryKey = resources.getString(resourceKey);
	    if (dictionaryKey.trim().length() == 0) {
	        throw new MissingResourceException(
	            "Resource "+resourceKey+" was blank",
	            "ResourceBundle", resourceKey);
	    }
        int entryNum = dictionary.lookup(dictionaryKey);
        if (entryNum >= 0) {
            DictionaryEntry entry = dictionary.getEntry(entryNum);
            if (entry.matches(dictionaryKey)) {
                return entry.makeSignLabel();
            }
        }
        throw new MissingResourceException(
            "Value \""+dictionaryKey+"\" for resource "+resourceKey+
            " does not match an entry in the dictionary",
            "ResourceBundle", resourceKey);
    }
    
    private void read() throws IOException {

        dictFile = new DictionaryFile(filename+".dic");
        DataInputStream in = new DataInputStream(
                             new BufferedInputStream(
                             new FileInputStream(filename+".din")));

		int oldAvail = in.available();
        if (in.read() != SWDICT_VERSION) {
            throw new IOException("Bad dictionary index file "+filename+".  Version is supposed to be "+SWDICT_VERSION);
        }
        for (int i = 0; i < 8; i++) {
            dheader[i] = PCNumbers.makeShort(in.read(),in.read());
        }

        dictFree = dheader[0];
        dictRecs = dheader[1];

        int country = PCNumbers.makeShort(in.read(),in.read());
		byte[] charMap = new byte[256];
		for (int i = 0; i < 256; i++) {
		    charMap[i] = in.readByte(); // 256 bytes
		}
		PairRec[] pairs = new PairRec[MAX_PAIR+1];
		for (int i = 0; i <= MAX_PAIR; i++) {
		    pairs[i] = new PairRec();
		    pairs[i].read(in); // 30 bytes
		}

		int numRead = oldAvail - in.available();

		if ((512 - numRead) != 207) {
		    throw new IOException("Expected to be 207 bytes from header but am "+(512-numRead));
		}
		in.skip(512 - numRead); // skip header

		// cool

        byte[] byteArray = new byte[9];
        numEntries = 0;
		for (int i = 0; i < NUM_VOLUMES; i++) {
            numEntries += dheader[i+2];
		}

        entries = new DictionaryEntry[numEntries];

        for (int j = 0; j < numEntries; j++) {
            int len = in.read();
            in.readFully(byteArray);
            String key = new String(byteArray,0,0,len);
            int recnum = PCNumbers.makeShort(in.read(),in.read());
            DictionaryEntry entry = new DictionaryEntry(dictFile, recnum, key);
            entries[j] = entry;
        }

        //if (in.available() != 0) {
        //    throw new IOException("Expected to have read entire dictionary index file, but have "+in.available()+" bytes left.");
        //}
        in.close();
    }

    public int lookup(String searchString) {
        if (failed) {
            return -1;
        }
        int rtn = 0;
        int low = 0;
        int high = entries.length - 1;
        int idx = -1;
        while (high > low) {
            int next = low + (high-low)/2;
            if (next == idx) {
                return rtn+next+1;
            }
            idx = next;
            DictionaryEntry test = entries[idx];
            int result = test.compareTo(searchString);
            if (result == 0) {
                return rtn+idx;
            } else if (result > 0) {
                high = idx;
            } else {
                low = idx;
            }
        }
        if (high < (entries.length-1)) {
            return rtn+high+1;
        }
        return -1;
    }
    
    public int getNumEntries() {
        return numEntries;
    }

    public DictionaryEntry getEntry(int entryNum) {
        if (!failed && entryNum < numEntries) {
            return entries[entryNum];
        }
        return null;
    }
}
